"""Functionality common to displaying dataset and variables metadata."""

from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any

from dash import dcc

from datadoc import state

if TYPE_CHECKING:
    from collections.abc import Callable

    from dash.development.base_component import Component
    from datadoc_model.model import LanguageStringType
    from pydantic import BaseModel

    from datadoc.frontend.callbacks.utils import MetadataInputTypes

logger = logging.getLogger(__name__)

INPUT_KWARGS = {
    "debounce": True,
    "style": {"width": "100%"},
    "className": "ssb-input",
}
NUMBER_KWARGS = dict(type="number", **INPUT_KWARGS)
DROPDOWN_KWARGS = {
    "style": {"width": "100%"},
    "className": "ssb-dropdown",
}


def kwargs_factory() -> dict[str, t.Any]:
    """Initialize the field extra_kwargs.

    We aren't allowed to directly assign a mutable type like a dict to
    a dataclass field.
    """
    return INPUT_KWARGS


def get_standard_metadata(metadata: BaseModel, identifier: str) -> MetadataInputTypes:
    """Get a metadata value from the model."""
    value = metadata.model_dump()[identifier]
    if value is None:
        return None
    return str(value)


def get_metadata_and_stringify(metadata: BaseModel, identifier: str) -> str:
    """Get a metadata value from the model and cast to string."""
    return str(get_standard_metadata(metadata, identifier))


def get_multi_language_metadata(metadata: BaseModel, identifier: str) -> str | None:
    """Get a metadata value supportng multiple languages from the model."""
    value: LanguageStringType = getattr(metadata, identifier)
    if value is None:
        return value
    return str(getattr(value, state.current_metadata_language))


def get_comma_separated_string(metadata: BaseModel, identifier: str) -> str:
    """Get a metadata value which is a list of strings from the model and convert it to a comma separated string."""
    value: list[str] = getattr(metadata, identifier)
    try:
        return ", ".join(value)
    except TypeError:
        return ""


@dataclass
class DisplayMetadata:
    """Controls for how a given metadata field should be displayed."""

    identifier: str
    display_name: str
    description: str
    obligatory: bool = False
    editable: bool = True
    url: bool = False
    multiple_language_support: bool = False


@dataclass
class DisplayVariablesMetadata(DisplayMetadata):
    """Controls for how a given metadata field should be displayed.

    Specific to variable fields.
    """

    options: dict[str, list[dict[str, str]]] | None = None
    presentation: str | None = "input"


@dataclass
class DisplayDatasetMetadata(DisplayMetadata):
    """Controls for how a given metadata field should be displayed.

    Specific to dataset fields.
    """

    extra_kwargs: dict[str, Any] = field(default_factory=kwargs_factory)
    component: type[Component] = dcc.Input
    value_getter: Callable[[BaseModel, str], Any] = get_standard_metadata



"""Global state.

DANGER: This global is safe when Datadoc is run as designed, with
an individual instance per user run within a Jupyter Notebook.

If Datadoc is redeployed as a multi-user web app then this storage
strategy must be modified, since users will modify each others data.
See here: https://dash.plotly.com/sharing-data-between-callbacks
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datadoc.backend.datadoc_metadata import DataDocMetadata
    from datadoc.enums import SupportedLanguages


# Global metadata container
metadata: DataDocMetadata

current_metadata_language: SupportedLanguages



"""Enumerations used in Datadoc."""
from __future__ import annotations

from enum import Enum

from datadoc_model import model
from datadoc_model.model import LanguageStringType


class SupportedLanguages(str, Enum):
    """The list of languages metadata may be recorded in.

    Reference: https://www.iana.org/assignments/language-subtag-registry/language-subtag-registry
    """

    NORSK_BOKMÅL = "nb"
    NORSK_NYNORSK = "nn"
    ENGLISH = "en"


class LanguageStringsEnum(Enum):
    """Enum class for storing LanguageStringType objects."""

    def __init__(
        self,
        language_strings: LanguageStringType,
    ) -> None:
        """Store the LanguageStringType object for displaying enum values in multiple languages.

        We don't particularly care what the value of the enum is,
        but when serialised it's convenient and readable to use the
        name of the enum, so we set the value to be the name.
        """
        self._value_ = self.name
        self.language_strings = language_strings

    @classmethod
    def _missing_(cls, value: object) -> LanguageStringsEnum:
        """Support constructing an enum member from a supplied name string."""
        try:
            member: LanguageStringsEnum = cls._member_map_[str(value)]  # type: ignore [assignment]
        except KeyError as e:
            # Raise the expected exception with a useful explanation
            message = f"{value} is not a valid {cls.__qualname__}"
            raise ValueError(message) from e
        else:
            return member

    def get_value_for_language(
        self,
        language: SupportedLanguages,
    ) -> str:
        """Retrieve the string for the relevant language."""
        return str(getattr(self.language_strings, language.value))


class Assessment(LanguageStringsEnum):
    """Sensitivity of data."""

    SENSITIVE = LanguageStringType(
        en=model.Assessment.SENSITIVE.value,
        nn="SENSITIV",
        nb="SENSITIV",
    )
    PROTECTED = LanguageStringType(
        en=model.Assessment.PROTECTED.value,
        nn="BESKYTTET",
        nb="BESKYTTET",
    )
    OPEN = LanguageStringType(en=model.Assessment.OPEN.value, nn="ÅPEN", nb="ÅPEN")


class DatasetStatus(LanguageStringsEnum):
    """Lifecycle status of a dataset."""

    DRAFT = LanguageStringType(
        en=model.DatasetStatus.DRAFT.value,
        nn="UTKAST",
        nb="UTKAST",
    )
    INTERNAL = LanguageStringType(
        en=model.DatasetStatus.INTERNAL.value,
        nn="INTERN",
        nb="INTERN",
    )
    EXTERNAL = LanguageStringType(
        en=model.DatasetStatus.EXTERNAL.value,
        nn="EKSTERN",
        nb="EKSTERN",
    )
    DEPRECATED = LanguageStringType(
        en=model.DatasetStatus.DEPRECATED.value,
        nn="UTGÅTT",
        nb="UTGÅTT",
    )


class DatasetState(LanguageStringsEnum):
    """Processing state of a dataset."""

    SOURCE_DATA = LanguageStringType(
        en=model.DatasetState.SOURCE_DATA.value,
        nn="KILDEDATA",
        nb="KILDEDATA",
    )
    INPUT_DATA = LanguageStringType(
        en=model.DatasetState.INPUT_DATA.value,
        nn="INNDATA",
        nb="INNDATA",
    )
    PROCESSED_DATA = LanguageStringType(
        en=model.DatasetState.PROCESSED_DATA.value,
        nn="KLARGJORTE DATA",
        nb="KLARGJORTE DATA",
    )
    STATISTICS = LanguageStringType(
        en=model.DatasetState.STATISTICS.value,
        nn="STATISTIKK",
        nb="STATISTIKK",
    )
    OUTPUT_DATA = LanguageStringType(
        en=model.DatasetState.OUTPUT_DATA.value,
        nn="UTDATA",
        nb="UTDATA",
    )


class UnitType(LanguageStringsEnum):
    """Statistical unit types.

    Ref: https://www.ssb.no/metadata/definisjoner-av-statistiske-enheter.
    """

    ARBEIDSULYKKE = LanguageStringType(
        en="WORK ACCIDENT",
        nn=model.UnitType.ARBEIDSULYKKE.value,
        nb=model.UnitType.ARBEIDSULYKKE.value,
    )
    BOLIG = LanguageStringType(
        en="HOUSING",
        nn=model.UnitType.BOLIG.value,
        nb=model.UnitType.BOLIG.value,
    )
    BYGNING = LanguageStringType(
        en="BUILDING",
        nn=model.UnitType.BYGNING.value,
        nb=model.UnitType.BYGNING.value,
    )
    EIENDOM = LanguageStringType(
        en="PROPERTY",
        nn=model.UnitType.EIENDOM.value,
        nb=model.UnitType.EIENDOM.value,
    )
    FAMILIE = LanguageStringType(
        en="FAMILY",
        nn=model.UnitType.FAMILIE.value,
        nb=model.UnitType.FAMILIE.value,
    )
    FORETAK = LanguageStringType(
        en="COMPANY",
        nn=model.UnitType.FORETAK.value,
        nb=model.UnitType.FORETAK.value,
    )
    FYLKE = LanguageStringType(
        en="REGION",
        nn=model.UnitType.FYLKE.value,
        nb=model.UnitType.FYLKE.value,
    )
    HAVNEANLOEP = LanguageStringType(
        en="PORT CALL",
        nn=model.UnitType.HAVNEANLOEP.value,
        nb=model.UnitType.HAVNEANLOEP.value,
    )
    HUSHOLDNING = LanguageStringType(
        en="HOUSEHOLD",
        nn=model.UnitType.HUSHOLDNING.value,
        nb=model.UnitType.HUSHOLDNING.value,
    )
    KJOERETOEY = LanguageStringType(
        en="VEHICLE",
        nn=model.UnitType.KJOERETOEY.value,
        nb=model.UnitType.KJOERETOEY.value,
    )
    KOMMUNE = LanguageStringType(
        en="COUNTY",
        nn=model.UnitType.KOMMUNE.value,
        nb=model.UnitType.KOMMUNE.value,
    )
    KURS = LanguageStringType(
        en="COURSE",
        nn=model.UnitType.KURS.value,
        nb=model.UnitType.KURS.value,
    )
    LOVBRUDD = LanguageStringType(
        en="CRIME",
        nn=model.UnitType.LOVBRUDD.value,
        nb=model.UnitType.LOVBRUDD.value,
    )
    PERSON = LanguageStringType(
        en="PERSON",
        nn=model.UnitType.PERSON.value,
        nb=model.UnitType.PERSON.value,
    )
    STAT = LanguageStringType(
        en="STATE",
        nn=model.UnitType.STAT.value,
        nb=model.UnitType.STAT.value,
    )
    STORFE = LanguageStringType(
        en="CATTLE",
        nn=model.UnitType.STORFE.value,
        nb=model.UnitType.STORFE.value,
    )
    TRAFIKKULYKKE = LanguageStringType(
        en="TRAFFIC ACCIDENT",
        nn=model.UnitType.TRAFIKKULYKKE.value,
        nb=model.UnitType.TRAFIKKULYKKE.value,
    )
    TRANSAKSJON = LanguageStringType(
        en="TRANSACTION",
        nn=model.UnitType.TRANSAKSJON.value,
        nb=model.UnitType.TRANSAKSJON.value,
    )
    VARE_TJENESTE = LanguageStringType(
        en="GOOD/SERVICE",
        nn=model.UnitType.VARE_TJENESTE.value,
        nb=model.UnitType.VARE_TJENESTE.value,
    )
    VERDIPAPIR = LanguageStringType(
        en="SERVICE",
        nn=model.UnitType.VERDIPAPIR.value,
        nb=model.UnitType.VERDIPAPIR.value,
    )
    VIRKSOMHET = LanguageStringType(
        en="BUSINESS",
        nn=model.UnitType.VIRKSOMHET.value,
        nb=model.UnitType.VIRKSOMHET.value,
    )


class TemporalityTypeType(LanguageStringsEnum):
    """Temporality of a dataset.

    More information about temporality type: https://statistics-norway.atlassian.net/l/c/HV12q90R
    """

    FIXED = LanguageStringType(
        en=model.TemporalityTypeType.FIXED.value,
        nn="FAST",
        nb="FAST",
    )
    STATUS = LanguageStringType(
        en=model.TemporalityTypeType.STATUS.value,
        nn="TVERRSNITT",
        nb="TVERRSNITT",
    )
    ACCUMULATED = LanguageStringType(
        en=model.TemporalityTypeType.ACCUMULATED.value,
        nn="AKKUMULERT",
        nb="AKKUMULERT",
    )
    EVENT = LanguageStringType(
        en=model.TemporalityTypeType.EVENT.value,
        nn="HENDELSE",
        nb="HENDELSE",
    )


class DataType(LanguageStringsEnum):
    """Simplified data types for metadata purposes."""

    STRING = LanguageStringType(en=model.DataType.STRING.value, nn="TEKST", nb="TEKST")
    INTEGER = LanguageStringType(
        en=model.DataType.INTEGER.value,
        nn="HELTALL",
        nb="HELTALL",
    )
    FLOAT = LanguageStringType(
        en=model.DataType.FLOAT.value,
        nn="DESIMALTALL",
        nb="DESIMALTALL",
    )
    DATETIME = LanguageStringType(
        en=model.DataType.DATETIME.value,
        nn="DATOTID",
        nb="DATOTID",
    )
    BOOLEAN = LanguageStringType(
        en=model.DataType.BOOLEAN.value,
        nn="BOOLSK",
        nb="BOOLSK",
    )


class VariableRole(LanguageStringsEnum):
    """The role of a variable in a dataset."""

    IDENTIFIER = LanguageStringType(
        en=model.VariableRole.IDENTIFIER.value,
        nn="IDENTIFIKATOR",
        nb="IDENTIFIKATOR",
    )
    MEASURE = LanguageStringType(
        en=model.VariableRole.MEASURE.value,
        nn="MÅLEVARIABEL",
        nb="MÅLEVARIABEL",
    )
    START_TIME = LanguageStringType(
        en=model.VariableRole.START_TIME.value,
        nn="STARTTID",
        nb="STARTTID",
    )
    STOP_TIME = LanguageStringType(
        en=model.VariableRole.STOP_TIME.value,
        nn="STOPPTID",
        nb="STOPPTID",
    )
    ATTRIBUTE = LanguageStringType(
        en=model.VariableRole.ATTRIBUTE.value,
        nn="ATTRIBUTT",
        nb="ATTRIBUTT",
    )



"""Perform operations on files stored with a range of technologies.

Defines a Protocol and the concrete implementations of it.
"""

from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING
from typing import Protocol
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

if TYPE_CHECKING:
    import os
    from io import IOBase
    from io import TextIOWrapper

GCS_PROTOCOL_PREFIX = "gs://"

logger = logging.getLogger(__name__)


class GCSObject:
    """Implementation of the Protocol 'StorageAdapter' for Google Cloud Storage."""

    def __init__(self, path: str | os.PathLike) -> None:
        """Initialize the class."""
        self._url = urlsplit(str(path))
        try:
            from dapla import AuthClient
            from dapla import FileClient

            if AuthClient.is_ready():
                # Running on Dapla, rely on dapla-toolbelt for auth
                self.fs = FileClient.get_gcs_file_system()
            else:
                # All other environments, rely on Standard Google credential system
                # If this doesn't work for you, try running the following commands:
                #
                # gcloud auth application-default revoke
                # gcloud auth application-default login
                from gcsfs import GCSFileSystem

                self.fs = GCSFileSystem()

        except ImportError as e:
            msg = "Missing support for GCS. Install datadoc with 'pip install ssb-datadoc[gcs]'"
            raise ImportError(msg) from e

    @staticmethod
    def for_path(path: str | os.PathLike) -> StorageAdapter:
        """Return an instance of this class instantiated for path."""
        return GCSObject(path)

    def _rebuild_url(self, new_path: str | os.PathLike) -> str:
        return urlunsplit(
            (self._url.scheme, self._url.netloc, str(new_path), None, None),
        )

    def open(self, **kwargs: str) -> IOBase:
        """Return a file-like-object."""
        return self.fs.open(self.location, **kwargs)

    def parent(self) -> str:
        """Return the logical parent of this object."""
        parent = pathlib.PurePosixPath(self._url.path).parent
        return self._rebuild_url(parent)

    def joinpath(self, part: str) -> None:
        """Join 'part' onto the current path.

        In-place operation.
        """
        self._url = urlsplit(
            self._rebuild_url(pathlib.Path(self._url.path) / part),
        )

    def exists(self) -> bool:
        """Return True if the object exists."""
        return self.fs.exists(self.location)

    def write_text(self, text: str) -> None:
        """Write the given text to disk."""
        f: TextIOWrapper
        with self.fs.open(self.location, mode="w") as f:
            f.write(text)

    @property
    def location(self) -> str:
        """Return a locator for this object."""
        return urlunsplit(self._url)


class LocalFile:
    """Implementation of the Protocol 'StorageAdapter' for file systems."""

    def __init__(self, path: str | os.PathLike) -> None:
        """Initialize the class."""
        self._path_object: pathlib.Path = pathlib.Path(path)

    @staticmethod
    def for_path(path: str | os.PathLike) -> StorageAdapter:
        """Return an instance of this class instantiated for path."""
        return LocalFile(path)

    def open(self, **kwargs: str) -> IOBase:
        """Return a file-like-object."""
        return pathlib.Path.open(self._path_object, **kwargs)  # type: ignore [call-overload]

    def parent(self) -> str:
        """Return the parent of this file."""
        return str(self._path_object.resolve().parent)

    def joinpath(self, part: str) -> None:
        """Join 'part' onto the current path.

        In-place operation.
        """
        self._path_object = self._path_object.joinpath(part)

    def exists(self) -> bool:
        """Return True if the file exists."""
        return self._path_object.exists()

    def write_text(self, text: str) -> None:
        """Write the given text to disk."""
        self._path_object.write_text(text, encoding="utf-8")

    @property
    def location(self) -> str:
        """Return a locator for this object."""
        return str(self._path_object)


class StorageAdapter(Protocol):
    """Implement this Protocol for the technologies on which we store datasets and metadata documents."""

    @staticmethod
    def for_path(path: str | os.PathLike) -> StorageAdapter:
        """Return a concrete class implementing this Protocol based on the structure of the path."""
        if str(path).startswith(GCS_PROTOCOL_PREFIX):
            return GCSObject(path)

        return LocalFile(path)

    def open(self, **kwargs: str) -> IOBase:
        """Return a file-like-object."""
        ...

    def parent(self) -> str:
        """Return the logical parent of this instance."""
        ...

    def joinpath(self, part: str) -> None:
        """Join 'part' onto the current path.

        In-place operation.
        """
        ...

    def exists(self) -> bool:
        """Return True if the object exists."""
        ...

    def write_text(self, text: str) -> None:
        """Write the given text to disk."""
        ...

    @property
    def location(self) -> str:
        """Return a locator for this object."""
        ...



"""Functionality for displaying variables metadata."""

from __future__ import annotations

import typing
from enum import Enum

from datadoc_model import model

from datadoc.frontend.fields.display_base import DisplayVariablesMetadata


class VariableIdentifiers(str, Enum):
    """As defined here: https://statistics-norway.atlassian.net/wiki/spaces/MPD/pages/3042869256/Variabelforekomst."""

    SHORT_NAME = "short_name"
    NAME = "name"
    DATA_TYPE = "data_type"
    VARIABLE_ROLE = "variable_role"
    DEFINITION_URI = "definition_uri"
    DIRECT_PERSON_IDENTIFYING = "direct_person_identifying"
    DATA_SOURCE = "data_source"
    POPULATION_DESCRIPTION = "population_description"
    COMMENT = "comment"
    TEMPORALITY_TYPE = "temporality_type"
    MEASUREMENT_UNIT = "measurement_unit"
    FORMAT = "format"
    CLASSIFICATION_URI = "classification_uri"
    SENTINEL_VALUE_URI = "sentinel_value_uri"
    INVALID_VALUE_DESCRIPTION = "invalid_value_description"
    IDENTIFIER = "id"
    CONTAINS_DATA_FROM = "contains_data_from"
    CONTAINS_DATA_UNTIL = "contains_data_until"


DISPLAY_VARIABLES = {
    VariableIdentifiers.SHORT_NAME: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.SHORT_NAME.value,
        display_name="Kortnavn",
        description="Fysisk navn på variabelen i datasettet. Bør tilsvare anbefalt kortnavn.",
        obligatory=True,
        editable=False,
    ),
    VariableIdentifiers.NAME: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.NAME.value,
        display_name="Navn",
        description="Variabelnavn kan arves fra VarDef, men kan også dokumenteres/endres her.",
        obligatory=True,
        multiple_language_support=True,
    ),
    VariableIdentifiers.DATA_TYPE: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.DATA_TYPE.value,
        display_name="Datatype",
        description="Datatype",
        obligatory=True,
        presentation="dropdown",
    ),
    VariableIdentifiers.VARIABLE_ROLE: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.VARIABLE_ROLE.value,
        display_name="Variabelens rolle",
        description="Variabelens rolle i datasett",
        obligatory=True,
        presentation="dropdown",
    ),
    VariableIdentifiers.DEFINITION_URI: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.DEFINITION_URI.value,
        display_name="Definition URI",
        description="En lenke (URI) til variabelens definisjon i SSB (Vardok/VarDef)",
        url=True,
        obligatory=True,
    ),
    VariableIdentifiers.DIRECT_PERSON_IDENTIFYING: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.DIRECT_PERSON_IDENTIFYING.value,
        display_name="DPI",
        description="Direkte personidentifiserende informasjon (DPI)",
        obligatory=True,
        presentation="dropdown",
    ),
    VariableIdentifiers.DATA_SOURCE: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.DATA_SOURCE.value,
        display_name="Datakilde",
        description="Datakilde. Settes på datasettnivå, men kan overstyres på variabelforekomstnivå.",
        multiple_language_support=True,
    ),
    VariableIdentifiers.POPULATION_DESCRIPTION: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.POPULATION_DESCRIPTION.value,
        display_name="Populasjonen",
        description="Populasjonen variabelen beskriver kan spesifiseres nærmere her. Settes på datasettnivå, men kan overstyres på variabelforekomstnivå.",
        multiple_language_support=True,
    ),
    VariableIdentifiers.COMMENT: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.COMMENT.value,
        display_name="Kommentar",
        description="Ytterligere presiseringer av variabeldefinisjon",
        multiple_language_support=True,
    ),
    VariableIdentifiers.TEMPORALITY_TYPE: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.TEMPORALITY_TYPE.value,
        display_name="Temporalitetstype",
        description="Temporalitetstype. Settes enten for variabelforekomst eller datasett. Se Temporalitet, hendelser og forløp.",
        presentation="dropdown",
    ),
    VariableIdentifiers.MEASUREMENT_UNIT: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.MEASUREMENT_UNIT.value,
        display_name="Måleenhet",
        description="Måleenhet. Eksempel: NOK eller USD for valuta, KG eller TONN for vekt. Se også forslag til SSBs måletyper/måleenheter.",
        multiple_language_support=True,
    ),
    VariableIdentifiers.FORMAT: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.FORMAT.value,
        display_name="Format",
        description="Verdienes format (fysisk format eller regulært uttrykk) i maskinlesbar form ifm validering. Dette kan benyttes som en ytterligere presisering av datatypen (dataType) i de tilfellene hvor dette er relevant. ",
    ),
    VariableIdentifiers.CLASSIFICATION_URI: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.CLASSIFICATION_URI.value,
        display_name="Kodeverkets URI",
        description="Lenke (URI) til gyldige kodeverk (klassifikasjon eller kodeliste) i KLASS",
        url=True,
    ),
    VariableIdentifiers.SENTINEL_VALUE_URI: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.SENTINEL_VALUE_URI.value,
        display_name="Spesialverdienes URI",
        description="En lenke (URI) til en oversikt over 'spesialverdier' som inngår i variabelen.",
        url=True,
    ),
    VariableIdentifiers.INVALID_VALUE_DESCRIPTION: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.INVALID_VALUE_DESCRIPTION.value,
        display_name="Ugyldige verdier",
        description="En beskrivelse av ugyldige verdier som inngår i variabelen dersom spesialverdiene ikke er tilstrekkelige eller ikke kan benyttes.",
        multiple_language_support=True,
    ),
    VariableIdentifiers.IDENTIFIER: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.IDENTIFIER.value,
        display_name="Unik ID",
        description="Unik SSB identifikator for variabelforekomsten i datasettet",
        obligatory=False,
        editable=False,
    ),
    VariableIdentifiers.CONTAINS_DATA_FROM: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.CONTAINS_DATA_FROM.value,
        display_name="Inneholder data f.o.m.",
        description="Variabelforekomsten i datasettet inneholder data fra og med denne dato.",
    ),
    VariableIdentifiers.CONTAINS_DATA_UNTIL: DisplayVariablesMetadata(
        identifier=VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
        display_name="Inneholder data t.o.m.",
        description="Variabelforekomsten i datasettet inneholder data til og med denne dato.",
    ),
}

MULTIPLE_LANGUAGE_VARIABLES_METADATA = [
    m.identifier for m in DISPLAY_VARIABLES.values() if m.multiple_language_support
]

URL_VARIABLES_METADATA = [m.identifier for m in DISPLAY_VARIABLES.values() if m.url]

DISPLAYED_DROPDOWN_VARIABLES_METADATA = [
    m.identifier for m in DISPLAY_VARIABLES.values() if m.presentation == "dropdown"
]

DISPLAYED_DROPDOWN_VARIABLES_TYPES = []

types = typing.get_type_hints(model.Variable)

for m in DISPLAY_VARIABLES.values():
    if m.presentation == "dropdown":
        field_type = typing.get_args(types[m.identifier])[0]
        DISPLAYED_DROPDOWN_VARIABLES_TYPES.append(field_type)


OBLIGATORY_VARIABLES_METADATA = [
    m.identifier for m in DISPLAY_VARIABLES.values() if m.obligatory and m.editable
]



"""Components and layout which are not inside a tab."""

from __future__ import annotations

import dash_bootstrap_components as dbc
from dash import dcc
from dash import html

from datadoc import state
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.utils import get_dataset_path
from datadoc.frontend.components.builders import build_ssb_button
from datadoc.utils import get_app_version

COLORS = {"dark_1": "#F0F8F9", "green_1": "#ECFEED", "green_4": "#00824D"}

header = dbc.CardBody(
    dbc.Row(
        children=[
            html.H1("DataDoc", className="ssb-title", style={"color": "white"}),
        ],
    ),
    style={"backgroundColor": COLORS["green_4"]},
)

progress_bar = dbc.CardBody(
    style={"padding": "4px"},
    children=[dbc.Progress(id="progress-bar", color=COLORS["green_4"], value=40)],
)


def build_language_dropdown() -> dbc.Row:
    """Build the language dropdown."""
    return dbc.CardBody(
        dbc.Row(
            [
                dbc.Col(
                    html.P(f"v{get_app_version()}", className="small"),
                    align="end",
                ),
                dbc.Col(
                    dcc.Dropdown(
                        id="language-dropdown",
                        searchable=False,
                        value=state.current_metadata_language.value,
                        className="ssb-dropdown",
                        options=[
                            {"label": i.name, "value": i.value}
                            for i in SupportedLanguages
                        ],
                    ),
                    align="center",
                    width="auto",
                ),
            ],
            justify="between",
        ),
    )


def build_controls_bar() -> dbc.CardBody:
    """Build the Controls Bar.

    This contains:
    - A text input to specify the path to a dataset
    - A button to open a dataset
    - A button to save metadata to disk
    """
    return dbc.CardBody(
        children=[
            dbc.Row(
                [
                    dbc.Col(
                        dbc.Row(
                            [
                                dbc.Col(
                                    dcc.Input(
                                        value=get_dataset_path(),
                                        size="50",
                                        placeholder="Sti til datasettet f.eks 'gs://my-bucket/my-dataset.parquet'",
                                        id="dataset-path-input",
                                    ),
                                    align="center",
                                    width="auto",
                                ),
                                dbc.Col(
                                    build_ssb_button(
                                        text="Åpne fil",
                                        icon_class="bi bi-folder2-open",
                                        button_id="open-button",
                                    ),
                                    width=2,
                                ),
                            ],
                        ),
                        width=6,
                    ),
                    dbc.Col(),
                    dbc.Col(
                        build_ssb_button(
                            text="Lagre metadata",
                            icon_class="bi bi-save",
                            button_id="save-button",
                        ),
                        width=2,
                    ),
                ],
                justify="between",
            ),
        ],
    )



"""All decorated callback functions should be defined here.

Implementations of the callback functionality should be in other functions (in other files), to enable unit testing.
"""


from __future__ import annotations

from typing import TYPE_CHECKING

from dash import ALL
from dash import Dash
from dash import Input
from dash import Output
from dash import State
from dash import ctx

from datadoc import state
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.dataset import accept_dataset_metadata_input
from datadoc.frontend.callbacks.dataset import change_language_dataset_metadata
from datadoc.frontend.callbacks.dataset import open_dataset_handling
from datadoc.frontend.callbacks.variables import accept_variable_metadata_input
from datadoc.frontend.callbacks.variables import (
    update_variable_table_dropdown_options_for_language,
)
from datadoc.frontend.callbacks.variables import update_variable_table_language
from datadoc.frontend.components.dataset_tab import DATASET_METADATA_INPUT
from datadoc.frontend.fields.display_dataset import DISPLAYED_DROPDOWN_DATASET_METADATA

if TYPE_CHECKING:
    from datadoc.frontend.callbacks.utils import MetadataInputTypes


def register_callbacks(app: Dash) -> None:
    """Define and register callbacks."""

    @app.callback(
        Output("progress-bar", "value"),
        Output("progress-bar", "label"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        Input("variables-table", "data"),
    )
    def callback_update_progress(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
        data: list[dict],  # noqa: ARG001 argument required by Dash
    ) -> tuple[int, str]:
        """Update the progress bar when new data is entered."""
        completion = state.metadata.percent_complete
        return completion, f"{completion}%"

    @app.callback(
        Output("saved-metadata-success", "is_open"),
        Input("save-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(n_clicks: int) -> bool:
        """Save the metadata document to disk."""
        if n_clicks and n_clicks > 0:
            # Write the final completion percentage to the model
            state.metadata.meta.percentage_complete = state.metadata.percent_complete
            state.metadata.write_metadata_document()
            return True

        return False

    @app.callback(
        *[
            Output(
                {
                    "type": DATASET_METADATA_INPUT,
                    "id": m.identifier,
                },
                "options",
            )
            for m in DISPLAYED_DROPDOWN_DATASET_METADATA
        ],
        Output(
            {"type": DATASET_METADATA_INPUT, "id": ALL},
            "value",
        ),
        Input("language-dropdown", "value"),
    )
    def callback_change_language_dataset_metadata(
        language: str,
    ) -> tuple[object, ...]:
        """Update dataset metadata values upon change of language."""
        return change_language_dataset_metadata(SupportedLanguages(language))

    @app.callback(
        Output("dataset-validation-error", "is_open"),
        Output("dataset-validation-explanation", "children"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
        )

    @app.callback(
        Output("variables-table", "data"),
        Output("variables-validation-error", "is_open"),
        Output("variables-validation-explanation", "children"),
        State("variables-table", "active_cell"),
        Input("variables-table", "data"),
        State("variables-table", "data_previous"),
        Input("language-dropdown", "value"),
        prevent_initial_call=True,
    )
    def callback_variable_table(
        active_cell: dict,
        data: list[dict],
        data_previous: list[dict],
        language: str,
    ) -> tuple[list[dict], bool, str]:
        """Update data in the variable table.

        Triggered upon:
        - New data enetered in the variable table.
        - Change of language.

        Will display an alert if validation fails.
        """
        if ctx.triggered_id == "language-dropdown":
            return update_variable_table_language(SupportedLanguages(language))

        return accept_variable_metadata_input(data, active_cell, data_previous)

    @app.callback(
        Output("variables-table", "dropdown"),
        Input("language-dropdown", "value"),
    )
    def callback_variable_table_dropdown_options(
        language: str,
    ) -> dict[str, dict[str, object]]:
        """Update the options in variable table dropdowns when the language changes."""
        language = SupportedLanguages(language)
        return update_variable_table_dropdown_options_for_language(language)

    @app.callback(
        Output("opened-dataset-success", "is_open"),
        Output("opened-dataset-error", "is_open"),
        Output("opened-dataset-error-explanation", "children"),
        Output("language-dropdown", "value"),  # Used to force reload of metadata
        Input("open-button", "n_clicks"),
        State("dataset-path-input", "value"),
    )
    def callback_open_dataset(
        n_clicks: int,
        dataset_path: str,
    ) -> tuple[bool, bool, str, str]:
        """Open a dataset.

        Shows an alert on success or failure.

        To trigger reload of data in the UI, we update the
        language dropdown. This is a hack and could be replaced
        by a more formal mechanism.
        """
        return open_dataset_handling(n_clicks, dataset_path)



"""Callbacks relating to datasets."""

from __future__ import annotations

import logging
import traceback

from pydantic import ValidationError

from datadoc import state
from datadoc.enums import (
    SupportedLanguages,  # noqa: TCH001 import is needed for docs build
)
from datadoc.frontend.callbacks.utils import MetadataInputTypes
from datadoc.frontend.callbacks.utils import find_existing_language_string
from datadoc.frontend.callbacks.utils import get_options_for_language
from datadoc.frontend.callbacks.utils import open_file
from datadoc.frontend.callbacks.utils import update_global_language_state
from datadoc.frontend.fields.display_dataset import DISPLAYED_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DISPLAYED_DROPDOWN_DATASET_ENUMS
from datadoc.frontend.fields.display_dataset import MULTIPLE_LANGUAGE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DatasetIdentifiers

logger = logging.getLogger(__name__)


def open_dataset_handling(
    n_clicks: int,
    file_path: str,
) -> tuple[bool, bool, str, str]:
    """Handle errors and other logic around opening a dataset file."""
    try:
        state.metadata = open_file(file_path)
    except FileNotFoundError:
        return (
            False,
            True,
            f"Filen '{file_path}' finnes ikke.",
            state.current_metadata_language.value,
        )
    except Exception as e:  # noqa: BLE001
        return (
            False,
            True,
            "\n".join(traceback.format_exception_only(type(e), e)),
            state.current_metadata_language.value,
        )
    if n_clicks and n_clicks > 0:
        return True, False, "", state.current_metadata_language.value

    return False, False, "", state.current_metadata_language.value


def process_keyword(value: str) -> list[str]:
    """Convert a comma separated string to a list of strings.

    e.g. 'a,b ,c' -> ['a', 'b', 'c']
    """
    return [item.strip() for item in value.split(",")]


def process_special_cases(
    value: MetadataInputTypes,
    metadata_identifier: str,
) -> MetadataInputTypes:
    """Pre-process metadata where needed.

    Some types of metadata need processing before being saved
    to the model. Handle these cases here, other values are
    returned unchanged.
    """
    updated_value: MetadataInputTypes
    if metadata_identifier == DatasetIdentifiers.KEYWORD.value and isinstance(
        value,
        str,
    ):
        updated_value = process_keyword(value)
    elif metadata_identifier in MULTIPLE_LANGUAGE_DATASET_METADATA and isinstance(
        value,
        str,
    ):
        updated_value = find_existing_language_string(
            state.metadata.meta.dataset,
            value,
            metadata_identifier,
        )
    else:
        updated_value = value

    # Other values get returned unchanged
    return updated_value


def accept_dataset_metadata_input(
    value: MetadataInputTypes,
    metadata_identifier: str,
) -> tuple[bool, str]:
    """Handle user inputs of dataset metadata values."""
    logger.debug(
        "Received updated value = %s for metadata_identifier = %s",
        value,
        metadata_identifier,
    )
    try:
        value = process_special_cases(value, metadata_identifier)
        # Update the value in the model
        setattr(
            state.metadata.meta.dataset,
            metadata_identifier,
            value,
        )
    except ValidationError as e:
        show_error = True
        error_explanation = f"`{e}`"
        logger.debug("Caught ValidationError:", exc_info=True)
    else:
        show_error = False
        error_explanation = ""
        logger.debug(
            "Successfully updated value = %s for metadata_identifier = %s",
            value,
            metadata_identifier,
        )

    return show_error, error_explanation


def update_dataset_metadata_language() -> list[MetadataInputTypes]:
    """Return new values for ALL the dataset metadata inputs.

    This allows editing of strings in the chosen language.
    """
    return [
        m.value_getter(state.metadata.meta.dataset, m.identifier)
        for m in DISPLAYED_DATASET_METADATA
    ]


def change_language_dataset_metadata(
    language: SupportedLanguages,
) -> tuple[object, ...]:
    """Change the language for the displayed dataset metadata.

    This is done in three steps:
    - Update the chosen language globally.
    - Update the language for dropdown options.
    - Update the language of all the metadata values.
    """
    update_global_language_state(language)
    return (
        *(
            get_options_for_language(language, e)
            for e in DISPLAYED_DROPDOWN_DATASET_ENUMS
        ),
        update_dataset_metadata_language(),
    )



"""Handle reading, updating and writing of metadata."""
from __future__ import annotations

import json
import logging
import os
import pathlib
import typing as t
import uuid
from typing import TYPE_CHECKING

from datadoc_model import model

from datadoc.backend.dataset_parser import DatasetParser
from datadoc.backend.model_backwards_compatibility import upgrade_metadata
from datadoc.backend.storage_adapter import StorageAdapter
from datadoc.enums import DatasetState
from datadoc.enums import SupportedLanguages
from datadoc.enums import VariableRole
from datadoc.frontend.fields import display_dataset
from datadoc.frontend.fields import display_variables
from datadoc.utils import calculate_percentage
from datadoc.utils import get_timestamp_now

if TYPE_CHECKING:
    from datetime import datetime

logger = logging.getLogger(__name__)

METADATA_DOCUMENT_FILE_SUFFIX = "__DOC.json"

PLACEHOLDER_USERNAME = "default_user@ssb.no"


class DataDocMetadata:
    """Handle reading, updating and writing of metadata."""

    def __init__(
        self,
        dataset_path: str | os.PathLike[str] | None = None,
        metadata_document_path: str | os.PathLike[str] | None = None,
    ) -> None:
        """Read in a dataset if supplied, otherwise naively instantiate the class."""
        self.dataset: pathlib.Path | None = None
        self.metadata_document: StorageAdapter | None = None
        self.container: model.MetadataContainer | None = None

        self.dataset_state: DatasetState | None = None
        self.short_name: str | None = None
        self.current_user: str | None = None
        self.meta: model.DatadocJsonSchema = model.DatadocJsonSchema(
            percentage_complete=0,
            dataset=model.Dataset(),
            variables=[],
        )

        self.variables_lookup: dict[str, model.Variable] = {}

        if metadata_document_path:
            # In this case the user has specified an independent metadata document for editing
            # without a dataset.
            self.metadata_document = StorageAdapter.for_path(metadata_document_path)
            self.extract_metadata_from_existing_document(self.metadata_document)

        elif dataset_path:
            self.dataset = pathlib.Path(dataset_path)
            # The short_name is set as the dataset filename without file extension
            self.short_name = pathlib.Path(
                self.dataset,
            ).stem
            self.metadata_document = StorageAdapter.for_path(
                StorageAdapter.for_path(self.dataset).parent(),
            )
            self.metadata_document.joinpath(
                self.short_name + METADATA_DOCUMENT_FILE_SUFFIX,
            )
            self.dataset_state = self.get_dataset_state(self.dataset)

            self.extract_metadata_from_files()

        try:
            self.current_user = os.environ["JUPYTERHUB_USER"]
        except KeyError:
            self.current_user = PLACEHOLDER_USERNAME
            logger.warning(
                "JUPYTERHUB_USER env variable not set, using %s as placeholder",
                self.current_user,
            )

    def get_dataset_state(
        self,
        dataset: pathlib.Path,
    ) -> DatasetState | None:
        """Use the path to attempt to guess the state of the dataset."""
        dataset_path_parts = set(dataset.parts)
        for state in DatasetState:
            # We assume that files are saved in the Norwegian language as specified by SSB.
            norwegian_dataset_state_path_part = state.get_value_for_language(
                SupportedLanguages.NORSK_BOKMÅL,
            ).lower()
            norwegian_dataset_state_path_part_variations = {
                norwegian_dataset_state_path_part.replace(" ", x) for x in ["-", "_"]
            }
            # Match on any of the variations anywhere in the path.
            if norwegian_dataset_state_path_part_variations.intersection(
                dataset_path_parts,
            ):
                return state

        return None

    @staticmethod
    def get_dataset_version(
        dataset_stem: str,
    ) -> str | None:
        """Find version information if exists in filename.

        eg. 'v1' in filename 'person_data_v1.parquet'
        """
        minimum_elements_in_file_name: t.Final[int] = 2
        minimum_characters_in_version_string: t.Final[int] = 2
        split_file_name = str(dataset_stem).split("_")
        if len(split_file_name) >= minimum_elements_in_file_name:
            last_filename_element = str(split_file_name[-1])
            if (
                len(last_filename_element) >= minimum_characters_in_version_string
                and last_filename_element[0:1] == "v"
                and last_filename_element[1:].isdigit()
            ):
                return last_filename_element[1:]
        return None

    def extract_metadata_from_files(self) -> None:
        """Read metadata from an existing metadata document.

        If no metadata document exists, create one from scratch by extracting metadata
        from the dataset file.
        """
        if self.metadata_document is not None and self.metadata_document.exists():
            self.extract_metadata_from_existing_document(self.metadata_document)
        elif self.dataset is not None:
            self.extract_metadata_from_dataset(self.dataset, self.short_name or "")

            self.meta.dataset.id = uuid.uuid4()

            # Set default values for variables where appropriate
            v: model.Variable
            for v in self.meta.variables:
                if v.variable_role is None:
                    v.variable_role = VariableRole.MEASURE
                if v.direct_person_identifying is None:
                    v.direct_person_identifying = False

        if not self.meta.dataset.id:
            self.meta.dataset.id = uuid.uuid4()

        self.variables_lookup = {v.short_name: v for v in self.meta.variables}

    def extract_metadata_from_existing_document(self, document: StorageAdapter) -> None:
        """There's an existing metadata document, so read in the metadata from that."""
        fresh_metadata = {}
        try:
            with document.open(mode="r", encoding="utf-8") as file:
                fresh_metadata = json.load(file)
            logger.info(
                "Opened existing metadata file %s",
                document.location,
            )

            if self.is_metadata_in_container_structure(fresh_metadata):
                self.container = model.MetadataContainer.model_validate_json(
                    json.dumps(fresh_metadata),
                )
                datadoc_metadata = fresh_metadata["datadoc"]
            else:
                datadoc_metadata = fresh_metadata

            datadoc_metadata = upgrade_metadata(
                datadoc_metadata,
            )

            self.meta = model.DatadocJsonSchema.model_validate_json(
                json.dumps(datadoc_metadata),
            )

        except json.JSONDecodeError:
            logger.warning(
                "Could not open existing metadata file %s. \
                    Falling back to collecting data from the dataset",
                document.location,
                exc_info=True,
            )

    def is_metadata_in_container_structure(
        self,
        metadata: dict,
    ) -> bool:
        """At a certain point a metadata 'container' was introduced.

        The container provides a structure for different 'types' of metadata, such as 'datadoc', 'pseudonymization' etc.
        This method returns True if the metadata is in the container structure, False otherwise.
        """
        return "datadoc" in metadata and "dataset" in metadata["datadoc"]

    def extract_metadata_from_dataset(
        self,
        dataset: pathlib.Path,
        short_name: str,
    ) -> None:
        """Obtain what metadata we can from the dataset itself.

        This makes it easier for the user by 'pre-filling' certain fields.
        Certain elements are dependent on the dataset being saved according to SSB's standard.
        """
        self.ds_schema: DatasetParser = DatasetParser.for_file(dataset)

        self.meta.dataset = model.Dataset(
            short_name=self.short_name,
            dataset_state=self.dataset_state,
            version=self.get_dataset_version(short_name),
            data_source_path=self.dataset,
            created_by=self.current_user,
        )

        self.meta.variables = self.ds_schema.get_fields()

    def write_metadata_document(self) -> None:
        """Write all currently known metadata to file."""
        timestamp: datetime = get_timestamp_now()
        if self.meta.dataset.metadata_created_date is None:
            self.meta.dataset.metadata_created_date = timestamp
        if self.meta.dataset.metadata_created_by is None:
            self.meta.dataset.metadata_created_by = self.current_user
        self.meta.dataset.metadata_last_updated_date = timestamp
        self.meta.dataset.metadata_last_updated_by = self.current_user

        if self.container:
            self.container.datadoc = self.meta
        else:
            self.container = model.MetadataContainer(datadoc=self.meta)

        if self.metadata_document:
            self.metadata_document.write_text(self.container.model_dump_json(indent=4))
            logger.info("Saved metadata document %s", self.metadata_document.location)
        else:
            msg = "No metadata document to save"
            raise ValueError(msg)

    @property
    def percent_complete(self) -> int:
        """The percentage of obligatory metadata completed.

        A metadata field is counted as complete when any non-None value is
        assigned. Used for a live progress bar in the UI, as well as being
        saved in the datadoc as a simple quality indicator.
        """
        num_all_fields = len(display_dataset.OBLIGATORY_DATASET_METADATA_IDENTIFIERS)
        num_set_fields = len(
            [
                k
                for k, v in self.meta.dataset.model_dump().items()
                if k in display_dataset.OBLIGATORY_DATASET_METADATA_IDENTIFIERS
                and v is not None
            ],
        )

        for variable in self.meta.variables:
            num_all_fields += len(display_variables.OBLIGATORY_VARIABLES_METADATA)
            num_set_fields += len(
                [
                    k
                    for k, v in variable.model_dump().items()
                    if k in display_variables.OBLIGATORY_VARIABLES_METADATA
                    and v is not None
                ],
            )

        return calculate_percentage(num_set_fields, num_all_fields)



"""Factory functions for different components are defined here."""

import re
from dataclasses import dataclass
from enum import Enum
from enum import auto

import dash_bootstrap_components as dbc
from dash import html


class AlertTypes(Enum):
    """Types of alerts."""

    SUCCESS = auto()
    WARNING = auto()


@dataclass
class AlertType:
    """Attributes of a concrete alert type."""

    alert_class_name: str
    color: str

    @staticmethod
    def get_type(alert_type: AlertTypes) -> "AlertType":
        """Get a concrete alert type based on the given enum values."""
        return ALERT_TYPES[alert_type]


ALERT_TYPES = {
    AlertTypes.WARNING: AlertType(
        alert_class_name="ssb-dialog warning",
        color="danger",
    ),
    AlertTypes.SUCCESS: AlertType(
        alert_class_name="ssb-dialog",
        color="success",
    ),
}


def build_ssb_styled_tab(label: str, content: dbc.Container) -> dbc.Tab:
    """Make a Dash Tab according to SSBs Design System."""
    return dbc.Tab(
        label=label,
        # Replace all whitespace with dashes
        tab_id=re.sub(r"\s+", "-", label.lower()),
        label_class_name="ssb-tabs navigation-item",
        label_style={"margin-left": "10px", "margin-right": "10px"},
        style={"padding": "4px"},
        children=content,
    )


def build_ssb_alert(
    alert_type: AlertTypes,
    alert_identifier: str,
    title: str,
    content_identifier: str,
) -> dbc.Alert:
    """Make a Dash Alert according to SSBs Design System."""
    alert = AlertType.get_type(alert_type)
    return dbc.Alert(
        id=alert_identifier,
        is_open=False,
        dismissable=True,
        fade=True,
        color=alert.color,
        duration=5000 if alert_type == AlertTypes.SUCCESS else None,
        children=[
            html.H5(
                title,
            ),
            html.P(
                id=content_identifier,
            ),
        ],
    )


def build_ssb_button(text: str, icon_class: str, button_id: str) -> dbc.Button:
    """Make a Dash Button according to SSBs Design System."""
    return dbc.Button(
        [
            html.I(
                className=icon_class,
                style={"padding-right": "10px"},
            ),
            f"   {text}",
        ],
        class_name="ssb-btn primary-btn",
        id=button_id,
    )



"""Functions which aren't directly called from a decorated callback."""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING
from typing import TypeAlias

from datadoc_model import model

from datadoc import enums
from datadoc import state
from datadoc.backend.datadoc_metadata import METADATA_DOCUMENT_FILE_SUFFIX
from datadoc.backend.datadoc_metadata import DataDocMetadata

if TYPE_CHECKING:
    from enum import Enum
    from pathlib import Path

    import pydantic

    from datadoc.enums import SupportedLanguages

logger = logging.getLogger(__name__)


DATADOC_DATASET_PATH_ENV_VAR = "DATADOC_DATASET_PATH"

MetadataInputTypes: TypeAlias = str | list[str] | int | float | bool | None


def update_global_language_state(language: SupportedLanguages) -> None:
    """Update global language state."""
    logger.debug("Updating language: %s", language.name)
    state.current_metadata_language = language


def get_language_strings_enum(enum: Enum) -> enums.LanguageStringsEnum:
    """Get the correct language strings enum for the given enum.

    We need multiple languages to display in the front end, but the model only defines a single language in the enums.
    """
    language_strings_enum: enums.LanguageStringsEnum = getattr(enums, enum.__name__)  # type: ignore [attr-defined]
    if not issubclass(language_strings_enum, enums.LanguageStringsEnum):  # type: ignore [arg-type]
        message = f"Expected {language_strings_enum} to be a subclass of LanguageStringsEnum, but is {type(language_strings_enum)}"
        raise TypeError(
            message,
        )
    return language_strings_enum


def get_options_for_language(
    language: SupportedLanguages,
    enum: Enum,
) -> list[dict[str, str]]:
    """Generate the list of options based on the currently chosen language."""
    return [
        {
            "label": i.get_value_for_language(language),
            "value": i.name,
        }
        for i in get_language_strings_enum(enum)  # type: ignore [attr-defined]
    ]


def find_existing_language_string(
    metadata_model_object: pydantic.BaseModel,
    value: str,
    metadata_identifier: str,
) -> model.LanguageStringType:
    """Get or create a LanguageStrings object and return it."""
    # In this case we need to set the string to the correct language code
    language_strings = getattr(metadata_model_object, metadata_identifier)
    if language_strings is None:
        # This means that no strings have been saved yet so we need to construct
        # a new LanguageStrings object
        language_strings = model.LanguageStringType(
            **{state.current_metadata_language.value: value},
        )
    else:
        # In this case there's an existing object so we save this string
        # to the current language
        setattr(language_strings, state.current_metadata_language.value, value)
    return language_strings


def get_dataset_path() -> str | Path | None:
    """Extract the path to the dataset from the potential sources."""
    if state.metadata.dataset is not None:
        return state.metadata.dataset
    path_from_env = os.getenv(DATADOC_DATASET_PATH_ENV_VAR)
    logger.info(
        "Dataset path from %s: '%s'",
        DATADOC_DATASET_PATH_ENV_VAR,
        path_from_env,
    )
    return path_from_env


def open_file(file_path: str | Path | None = None) -> DataDocMetadata:
    """Load the given dataset into a DataDocMetadata instance."""
    if file_path and str(file_path).endswith(METADATA_DOCUMENT_FILE_SUFFIX):
        logger.info("Opening existing metadata document %s", file_path)
        return DataDocMetadata(metadata_document_path=file_path)

    dataset = file_path or get_dataset_path()
    logger.info("Opening dataset %s", dataset)
    return DataDocMetadata(dataset_path=dataset)



"""Components for different types of alerts are defined here."""

from __future__ import annotations

from datadoc.frontend.components.builders import AlertTypes
from datadoc.frontend.components.builders import build_ssb_alert

dataset_validation_error = build_ssb_alert(
    AlertTypes.WARNING,
    "dataset-validation-error",
    "Validering feilet",
    "dataset-validation-explanation",
)

variables_validation_error = build_ssb_alert(
    AlertTypes.WARNING,
    "variables-validation-error",
    "Validering feilet",
    "variables-validation-explanation",
)

opened_dataset_error = build_ssb_alert(
    AlertTypes.WARNING,
    "opened-dataset-error",
    "Kunne ikke åpne datasettet",
    "opened-dataset-error-explanation",
)

saved_metadata_success = build_ssb_alert(
    AlertTypes.SUCCESS,
    "saved-metadata-success",
    "Lagret metadata",
    "saved-metadata-success-explanation",
)

opened_dataset_success = build_ssb_alert(
    AlertTypes.SUCCESS,
    "opened-dataset-success",
    "Åpnet datasett",
    "opened-dataset-success-explanation",
)



"""Top-level entrypoint, configuration and layout for the datadoc app.

Members of this module should not be imported into any sub-modules, this will cause circular imports.
"""
from __future__ import annotations

import logging
import os
from pathlib import Path

import dash_bootstrap_components as dbc
from dash import Dash
from flask_healthz import healthz

from datadoc import state
from datadoc.backend.datadoc_metadata import DataDocMetadata
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.register_callbacks import register_callbacks
from datadoc.frontend.components.alerts import dataset_validation_error
from datadoc.frontend.components.alerts import opened_dataset_error
from datadoc.frontend.components.alerts import opened_dataset_success
from datadoc.frontend.components.alerts import saved_metadata_success
from datadoc.frontend.components.alerts import variables_validation_error
from datadoc.frontend.components.control_bars import build_controls_bar
from datadoc.frontend.components.control_bars import build_language_dropdown
from datadoc.frontend.components.control_bars import header
from datadoc.frontend.components.control_bars import progress_bar
from datadoc.frontend.components.dataset_tab import build_dataset_tab
from datadoc.frontend.components.variables_tab import build_variables_tab
from datadoc.utils import get_app_version
from datadoc.utils import pick_random_port
from datadoc.utils import running_in_notebook

logger = logging.getLogger(__name__)

NAME = "Datadoc"
DEFAULT_PORT: int = 7002
JUPYTERHUB_SERVICE_PREFIX_ENV = "JUPYTERHUB_SERVICE_PREFIX"


def build_app(app: type[Dash]) -> Dash:
    """Define the layout, register callbacks."""
    app.layout = dbc.Container(
        style={"padding": "4px"},
        children=[
            header,
            progress_bar,
            build_controls_bar(),
            variables_validation_error,
            dataset_validation_error,
            opened_dataset_error,
            saved_metadata_success,
            opened_dataset_success,
            dbc.CardBody(
                style={"padding": "4px"},
                children=[
                    dbc.Tabs(
                        id="tabs",
                        class_name="ssb-tabs",
                        children=[
                            build_dataset_tab(),
                            build_variables_tab(),
                        ],
                    ),
                ],
            ),
            build_language_dropdown(),
        ],
    )

    register_callbacks(app)

    return app


def get_app(dataset_path: str | None = None) -> tuple[Dash, int]:
    """Centralize all the ugliness around initializing the app."""
    logging.basicConfig(level=logging.INFO, force=True)
    logger.info("Datadoc version v%s", get_app_version())
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    state.metadata = DataDocMetadata(dataset_path)

    # This must be set to run correctly on Dapla Jupyter
    if JUPYTERHUB_SERVICE_PREFIX_ENV in os.environ:
        port = pick_random_port()
        requests_pathname_prefix = (
            f"{os.getenv(JUPYTERHUB_SERVICE_PREFIX_ENV, '/')}proxy/{port}/"
        )
    else:
        port = DEFAULT_PORT
        requests_pathname_prefix = "/"

    app = Dash(
        name=NAME,
        title=NAME,
        assets_folder=f"{Path(__file__).parent}/assets",
        requests_pathname_prefix=requests_pathname_prefix,
    )
    app = build_app(app)
    app.server.register_blueprint(healthz, url_prefix="/healthz")
    app.server.config["HEALTHZ"] = {
        "live": lambda: True,
        "ready": lambda: True,
        "startup": lambda: True,
    }
    logger.info("Built app with endpoints configured on /healthz")

    return app, port


def main(dataset_path: str | None = None) -> None:
    """Entrypoint when running as a script."""
    logging.basicConfig(level=logging.DEBUG, force=True)
    logger.info("Starting app with dataset_path = %s", dataset_path)
    app, port = get_app(dataset_path)
    if running_in_notebook():
        logger.info("Running in notebook")
        app.run(
            jupyter_mode="tab",
            jupyter_server_url=os.getenv("JUPYTERHUB_HTTP_REFERER", None),
            jupyter_height=1000,
            port=port,
        )
    else:
        # Assume running in server mode is better (largely for development purposes)
        logging.basicConfig(level=logging.DEBUG, force=True)
        logger.debug("Starting in development mode")
        app.run(debug=True, port=port)


if __name__ == "__main__":
    main()



"""Components and layout for the Dataset metadata tab."""

from __future__ import annotations

import dash_bootstrap_components as dbc
from dash import html

from datadoc.frontend.components.builders import build_ssb_styled_tab
from datadoc.frontend.fields.display_dataset import NON_EDITABLE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import OBLIGATORY_EDITABLE_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import OPTIONAL_DATASET_METADATA
from datadoc.frontend.fields.display_dataset import DisplayDatasetMetadata

DATASET_METADATA_INPUT = "dataset-metadata-input"


def build_dataset_metadata_accordion_item(
    title: str,
    metadata_inputs: list[DisplayDatasetMetadata],
) -> dbc.AccordionItem:
    """Build a Dash AccordionItem for the given Metadata inputs.

    Typically used to categorize metadata fields.
    """
    return dbc.AccordionItem(
        title=title,
        children=[
            dbc.Row(
                [
                    dbc.Col(html.Label(i.display_name)),
                    dbc.Col(
                        i.component(
                            placeholder=i.description,
                            disabled=not i.editable,
                            id={
                                "type": DATASET_METADATA_INPUT,
                                "id": i.identifier,
                            },
                            **i.extra_kwargs,
                        ),
                        width=5,
                    ),
                    dbc.Col(width=4),
                ],
            )
            for i in metadata_inputs
        ],
    )


def build_dataset_tab() -> dbc.Tab:
    """Build the Dataset metadata tab."""
    return build_ssb_styled_tab(
        "Datasett",
        dbc.Container(
            [
                dbc.Row(html.H2("Datasett detaljer", className="ssb-title")),
                dbc.Accordion(
                    always_open=True,
                    children=[
                        build_dataset_metadata_accordion_item(
                            "Obligatorisk",
                            OBLIGATORY_EDITABLE_DATASET_METADATA,
                        ),
                        build_dataset_metadata_accordion_item(
                            "Valgfritt",
                            OPTIONAL_DATASET_METADATA,
                        ),
                        build_dataset_metadata_accordion_item(
                            "Maskingenerert",
                            NON_EDITABLE_DATASET_METADATA,
                        ),
                    ],
                ),
            ],
        ),
    )



"""Callback functions to do with variables metadata."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pydantic import ValidationError

from datadoc import state
from datadoc.enums import SupportedLanguages
from datadoc.frontend.callbacks.utils import MetadataInputTypes
from datadoc.frontend.callbacks.utils import find_existing_language_string
from datadoc.frontend.callbacks.utils import get_options_for_language
from datadoc.frontend.fields.display_variables import (
    DISPLAYED_DROPDOWN_VARIABLES_METADATA,
)
from datadoc.frontend.fields.display_variables import DISPLAYED_DROPDOWN_VARIABLES_TYPES
from datadoc.frontend.fields.display_variables import (
    MULTIPLE_LANGUAGE_VARIABLES_METADATA,
)
from datadoc.frontend.fields.display_variables import VariableIdentifiers
from datadoc.utils import get_display_values

if TYPE_CHECKING:
    from datadoc_model import model

logger = logging.getLogger(__name__)


def get_boolean_options_for_language(
    language: SupportedLanguages,
) -> list[dict[str, bool | str]]:
    """Get boolean options for the given language.

    The Dash Datatable has no good support for boolean
    choices, so we use a Dropdown.
    """
    true_labels = {
        SupportedLanguages.ENGLISH: "Yes",
        SupportedLanguages.NORSK_NYNORSK: "Ja",
        SupportedLanguages.NORSK_BOKMÅL: "Ja",
    }
    false_labels = {
        SupportedLanguages.ENGLISH: "No",
        SupportedLanguages.NORSK_NYNORSK: "Nei",
        SupportedLanguages.NORSK_BOKMÅL: "Nei",
    }
    return [
        {
            "label": f"{true_labels[language]}",
            "value": True,
        },
        {
            "label": f"{false_labels[language]}",
            "value": False,
        },
    ]


def get_metadata_field(
    data: list[dict],
    data_previous: list[dict],
    active_cell: dict,
) -> str:
    """Find which field (column in the data table) has been updated."""
    for i in range(len(data)):
        # Find which column we're in; by diffing the current and previous data
        update_diff = list(data[i].items() - data_previous[i].items())
        if update_diff:
            return update_diff[-1][0]

    # When diffing fails, we fall back to the active cell (this happens
    # when the user pastes a value into the Data Table)
    return active_cell["column_id"]


def handle_multi_language_metadata(
    metadata_field: str,
    new_value: MetadataInputTypes,
    updated_row_id: str,
) -> MetadataInputTypes | model.LanguageStringType:
    """Handle updates to fields which support multiple languages."""
    if new_value is None:
        # This edge case occurs when the user removes the text in an input field
        # We want to ensure we only remove the content for the current language,
        # not create a new blank object!
        return find_existing_language_string(
            state.metadata.variables_lookup[updated_row_id],
            "",
            metadata_field,
        )

    if isinstance(new_value, str):
        return find_existing_language_string(
            state.metadata.variables_lookup[updated_row_id],
            new_value,
            metadata_field,
        )

    return new_value


def accept_variable_metadata_input(
    data: list[dict],
    active_cell: dict,
    data_previous: list[dict],
) -> tuple[list[dict], bool, str]:
    """Validate and save the value when variable metadata is updated."""
    show_error = False
    error_explanation = ""
    output_data = data
    metadata_field: str = get_metadata_field(data, data_previous, active_cell)

    for row_index in range(len(data)):
        # Update all the variables for this column to ensure we read in the value
        new_value: MetadataInputTypes = data[row_index][metadata_field]
        updated_row_id = data[row_index][VariableIdentifiers.SHORT_NAME.value]

        try:
            if metadata_field in MULTIPLE_LANGUAGE_VARIABLES_METADATA:
                new_value = handle_multi_language_metadata(
                    metadata_field,
                    new_value,
                    updated_row_id,
                )
            elif new_value == "":
                # Allow clearing non-multiple-language text fields
                new_value = None

            # Write the value to the variables structure
            setattr(
                state.metadata.variables_lookup[updated_row_id],
                metadata_field,
                new_value,
            )
        except ValidationError as e:
            show_error = True
            error_explanation = f"`{e}`"
            output_data = data_previous
            logger.debug("Caught ValidationError:", exc_info=True)
        else:
            logger.debug("Successfully updated %s with %s", updated_row_id, new_value)

    return output_data, show_error, error_explanation


def update_variable_table_dropdown_options_for_language(
    language: SupportedLanguages,
) -> dict[str, dict[str, object]]:
    """Retrieve enum options for dropdowns in the Datatable.

    Handles the special case of boolean values which we represent in the Datatable
    with a Dropdown but they're not backed by an Enum. Example return data structure as follows:

    ..  code-block:: python

        {
            "data_type": {
                "options": [
                    {"label": "TEKST", "value": "STRING"},
                    {"label": "HELTALL", "value": "INTEGER"},
                    {"label": "DESIMALTALL", "value": "FLOAT"},
                    {"label": "DATOTID", "value": "DATETIME"},
                    {"label": "BOOLSK", "value": "BOOLEAN"},
                ],
            },
            "direct_person_identifying": {
                "options": [
                    {"label": "Ja", "value": True},
                    {"label": "Nei", "value": False},
                ],
            },
            "temporality_type": {"options": [{"label": "FAST", "value": "FIXED"}]},
        }

    Args:
        language (SupportedLanguages): The language for metadata entry selected by the user.

    Returns:
        Data structure with options for all dropdowns.

    """
    options: list[dict[str, object]] = []
    for field_type in DISPLAYED_DROPDOWN_VARIABLES_TYPES:
        value = (
            get_boolean_options_for_language(language)
            if field_type is bool
            else get_options_for_language(language, field_type)
        )
        options.append({"options": value})
    return dict(zip(DISPLAYED_DROPDOWN_VARIABLES_METADATA, options, strict=True))


def update_variable_table_language(
    language: SupportedLanguages,
) -> tuple[list[dict], bool, str]:
    """Get data in the relevant language."""
    state.current_metadata_language = language
    logger.debug("Updated variable table language: %s", language.name)
    return (
        [
            get_display_values(
                v,
                state.current_metadata_language,
            )
            for v in state.metadata.meta.variables
        ],
        False,  # Don't show validation error
        "",  # No validation explanation needed
    )



"""Components and layout for the Variables metadata tab."""

from __future__ import annotations

import dash_bootstrap_components as dbc
from dash import dash_table
from dash import html

from datadoc import state
from datadoc.frontend.components.builders import build_ssb_styled_tab
from datadoc.frontend.fields.display_variables import DISPLAY_VARIABLES
from datadoc.frontend.fields.display_variables import VariableIdentifiers
from datadoc.utils import get_display_values


def build_variables_tab() -> dbc.Tab:
    """Build the Variables metadata tab."""
    return build_ssb_styled_tab(
        "Variabler",
        dbc.Container(
            children=[
                dbc.Row(html.H2("Variabel detaljer", className="ssb-title")),
                dbc.Row(
                    dash_table.DataTable(
                        id="variables-table",
                        # Populate fields with known values
                        data=[
                            get_display_values(v, state.current_metadata_language)
                            for v in state.metadata.meta.variables
                        ],
                        # Define columns based on the information in DISPLAY_VARIABLES
                        columns=[
                            {
                                "name": variable.display_name,
                                "id": variable.identifier,
                                "editable": variable.editable,
                                "presentation": variable.presentation,
                                "hideable": variable.editable,
                            }
                            for variable in DISPLAY_VARIABLES.values()
                            if variable.identifier
                            != VariableIdentifiers.IDENTIFIER.value  # Should be removed from the model, for now we hide it
                        ],
                        # Non-obligatory variables are hidden by default
                        hidden_columns=[
                            v.identifier
                            for v in DISPLAY_VARIABLES.values()
                            if v.obligatory is False
                        ],
                        # Include tooltips for all columns
                        tooltip_header={
                            v.identifier: v.description
                            for v in DISPLAY_VARIABLES.values()
                        },
                        editable=True,
                        # Exclude short_name column from scrolling
                        fixed_columns={"headers": True, "data": 1},
                        # Enable sorting and pagination
                        sort_action="native",
                        page_action="native",
                        page_size=20,
                        # Enable filtering
                        filter_action="native",
                        filter_options={"case": "insensitive"},
                        # Use horizontal scroll, keep full width
                        style_table={
                            "overflowX": "auto",
                            "minWidth": "100%",
                            "accentColor": "black",
                        },
                    ),
                ),
            ],
        ),
    )
